(* Lexic *)
{
open Syntax

exception Parse_error of int * int * string

let lexical_error l s = raise (Parse_error(Lexing.lexeme_start l, Lexing.lexeme_end l,s))

let keywords = [
  "missing", MISSING;
  "the", THE;
  "and", AND;
  "broken", BROKEN;
  "it", IT;
  "in", IN;
  "is", IS;
  "a", A;
  "pristine", PRISTINE;
  "condition", CONDITION;
];;

let keywords_table =
  let h = Hashtbl.create 64 in
  List.iter (fun (kwd,tok) -> Hashtbl.add h kwd tok) keywords;
  h
;;

(*** colors *)
let colors = [|
    "azure";
    "beige";
    "bisque";
    "blaze-orange";
    "bondi-blue";
    "brass";
    "bright-green";
    "bright-turquoise";
    "bright-violet";
    "bronze";
    "brown";
    "burlywood-colored";
    "burnt-orange";
    "carrot";
    "celadon";
    "cerise";
    "chartreuse";
    "chocolate-colored";
    "cobalt";
    "cream";
    "crimson";
    "cyan";
    "deep-pink";
    "deep-sky-blue";
    "denim";
    "dim-gray";
    "discounted";
    "dodger-blue";
    "eggplant";
    "emerald";
    "exceptional";
    "fern-green";
    "flax";
    "floating";
    "floral-white";
    "foreign";
    "forest-green";
    "fuchsia";
    "gray20";
    "gray30";
    "gray40";
    "gray50";
    "gray60";
    "gray70";
    "gray90";
    "gray-tea-green";
    "green";
    "green-yellow";
    "heavy";
    "heliotrope";
    "hot-pink";
    "imaginary";
    "imported";
    "indigo";
    "ivory";
    "jade";
    "khaki";
    "lavender";
    "left-handed";
    "lime";
    "lime-green";
    "magenta";
    "malachite";
    "midnight-blue";
    "miniature";
    "mint-green";
    "misty-rose";
    "moccasin";
    "moss-green";
    "navy-blue";
    "old";
    "old-gold";
    "olive-drab";
    "olive-green";
    "orange-red";
    "organic";
    "pale-brown";
    "pale-carmine";
    "pale-chestnut";
    "pale-cornflower-blue";
    "pale-green";
    "pastel-pink";
    "peach-orange";
    "peach-yellow";
    "persian-blue";
    "pine-green";
    "plum";
    "powder-blue";
    "prussian-blue";
    "puce";
    "pumpkin";
    "purple";
    "raw-umber";
    "red";
    "reverse-chirality";
    "robin-egg-blue";
    "rosy-brown";
    "rotating";
    "royal-blue";
    "sangria";
    "scarlet";
    "school-bus-yellow";
    "sea-green";
    "seashell";
    "sepia";
    "shiny";
    "sienna";
    "slate-blue";
    "slate-gray";
    "snow";
    "steel-blue";
    "tangerine";
    "taupe";
    "tepid";
    "thistle-colored";
    "ultramarine";
    "vegan";
    "vermilion";
    "violet";
    "violet-eggplant";
    "violet-red";
    "viridian";
    "water-logged";
    "wheat-colored";
    "yellow-green";
|];;
(* ***)

let colors_table =
  let m = Array.length colors in
  let h = Hashtbl.create m in
  for i = 0 to m - 1 do
    Hashtbl.add h colors.(i) i
  done;
  h
;;
}
let blancs = [' ' '\n' '\t' '\r']+
let upper = ['A'-'Z']
let alpha = ['a'-'z''A'-'Z']
let alphanumeric = alpha | ['0'-'9']
let decimal = ['0'-'9']
let sign = '+'|'-'
let space = [' ''\t''\r''\n']
rule token = parse
(* Operators *)
(* Various non-alphabetic symbols *)
| "(" { LPAREN }
| ")" { RPAREN }
| "," { COMMA }
| ":" { COLON }
| ">:" { PROMPT }
| "." { DOT }
| "(broken)" { L_BROKEN_R }
| (upper "-" decimal decimal decimal decimal "-" upper upper upper) as w { PART w }
| (alpha | (alphanumeric (alphanumeric|"-")* alphanumeric)) as u {
  let u = String.lowercase u in
  try
    COLOR(Hashtbl.find colors_table u)
  with
  | Not_found ->
      try
        Hashtbl.find keywords_table u
      with
      | Not_found ->
          BLA u
}
| space+ { token lexbuf }
| eof { EOF }
| _ as c { token lexbuf }
{
(* Epilogue. *)
}
